#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Foobar.py: Description of what foobar does."""

# import pytest

__author__ = "Fabrice Allain"
__copyright__ = "Institut Curie 2020"


def test_main():
    """Check if we can instantiate a GConfor object with a folder"""
    pass
